<?php

namespace PixelRockstar\Api;

use PixelRockstar\Media\Image;
use PixelRockstar\Media\Item;
use PixelRockstar\User\Subscription;
use PixelRockstar\User\User;

class Api
{
    const PIXELROCKSTAR_API_URL = "https://api2.pixelrockstar.com/";

    const API_USER_AGENT = 'PixelRockstar WP plugin';

    const RESPONSE_OK = "success";
    const RESPONSE_FAILED = "failure";

    const NOTIFY_FILE_DOWNLOAD_OK = "download_ok";
    const NOTIFY_FILE_DOWNLOAD_FAILED = "download_failed";
    const NOTIFY_FILE_DOWNLOAD_BAD_CHECKSUM = "bad_checksum";

    const DEFAULT_CACHE_TIME = 300;

    protected $cacheTtl;

    /** @var Credentials */
    private $credentials;

    public function __construct(Credentials $credentials, $cacheTtl = self::DEFAULT_CACHE_TIME)
    {
        $this->credentials = $credentials;
        $this->cacheTtl = $cacheTtl;
    }

    /**
     * @param string $url
     * @param array  $params
     *
     * @return resource
     */
    private function initCurl($url, $params = array())
    {
        $params["pixelrockstar_debug"] = array(
            "plugin_version" => PIXELROCKSTAR_VERSION,
            "wp_version" => get_bloginfo("version")
        );

        $ch = curl_init();
        $options = array(
            CURLOPT_URL => $url,
            CURLOPT_HTTPAUTH => CURLAUTH_BASIC,
            CURLOPT_USERPWD => $this->credentials->getUsername() . ':' . $this->credentials->getPassword(),
            CURLOPT_USERAGENT => self::API_USER_AGENT,
            CURLOPT_SSL_VERIFYPEER => false,
            CURLOPT_HEADER => false,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_FOLLOWLOCATION => false,
            CURLOPT_POST => $params,
            CURLOPT_POSTFIELDS => build_query($params),
            CURLOPT_REFERER => get_site_url() // for troubleshoting reasons
        );

        curl_setopt_array($ch, $options);

        return $ch;
    }

    /**
     * Callback function for parsing incoming curl headers
     *
     * @param resource $ch  Curl resource
     * @param string   $str Received header data
     *
     * @return int  Size of received data
     */
    private static function curlHeaderCallback($ch, $str)
    {
        // If call returns 404, do not save response into target file
        if (stripos($str, "404 Not Found") !== false) {
            curl_setopt($ch, CURLOPT_FILE, null);
            curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        }

        return strlen($str);
    }

    /**
     * @param string $url
     * @param array  $params
     *
     * @return \stdClass
     *
     * @throws ApiException
     */
    private function performApiCall($url, array $params = array())
    {
        $ch = self::initCurl($url, $params);

        $data = curl_exec($ch);
        curl_close($ch);

        $json = json_decode($data);
        if (!$json) {
            throw new ApiException("Invalid response from Pixelrockstar API server. Please try again later.");
        }

        if (!empty($json->api->status) && $json->api->status == self::RESPONSE_FAILED) {
            throw new ApiException($json->api->message, @$json->api->error_code ?: 0);
        }

        return $json;
    }

    /**
     * @param string $method
     * @param array  $params
     * @param bool   $cache
     *
     * @return \stdClass
     * @throws ApiException
     */
    public function get($method, array $params = array(), $cache = true)
    {
        $cache = false;
        $data = false;
        $url = self::PIXELROCKSTAR_API_URL . $method . ".json";

        $hash = md5($url . serialize($params));

        if ($cache && $this->cacheTtl) {
            $data = get_transient($hash);
        }

        if ($data === false) {
            $data = $this->performApiCall($url, $params);
            if ($cache && $this->cacheTtl) {
                set_transient($hash, $data, $this->cacheTtl);
            }
        }

        return $data;
    }

    /**
     * Get instance of `PixelRockstarUser` populated with current user's data
     *
     * @return User
     */
    public function getUser()
    {
        $user = new User();

        try {
            $data = $this->get('user-info', array(), false);
            if (empty($data->username)) {
                throw new \Exception("Empty user data");
            }
        } catch (\Exception $e) {
            return $user;
        }

        $user->setUserId($data->userId);
        $user->setUsername($data->username);
        $user->setType($data->type);
        $user->setEmail($data->email);
        $user->setFirstname($data->firstname);
        $user->setLastname($data->lastname);
        $user->setCountry($data->country);
        $user->setPhone($data->phone);
        $user->setCity($data->city);
        $user->setCompany($data->company);
        $user->setPostalCode($data->postalCode);
        $user->setState($data->state);
        $user->setStreet($data->street);
        $user->setStreet2($data->street2);
        $user->setStreet3($data->street3);
        $user->setVatId($data->vatId);

        $subscription = new Subscription();
        if (!empty($data->subscription)) {
            $subscription->setDownloadsInPackage($data->subscription->downloadsInPackage);
            $subscription->setDownloadsUsed($data->subscription->downloadsUsed);
            $subscription->setDownloadsAvailable($data->subscription->downloadsAvailable);
            $subscription->setCreditsAvailable($data->subscription->creditsAvailable);
            if ($data->subscription->startDate > 0) {
                $subscription->setStartDate(\DateTime::createFromFormat("U", $data->subscription->startDate));
            }
            if ($data->subscription->endDate > 0) {
                $subscription->setEndDate(\DateTime::createFromFormat("U", $data->subscription->endDate));
            }
            $subscription->setRenewalPeriod($data->subscription->renewalPeriod);
        }

        $user->setSubscription($subscription);

        return $user;
    }

    /**
     * Updates user details
     *
     * @param User $user
     *
     * @return bool
     */
    public function updateUser(User $user)
    {
        $data = array(
            "email" => $user->getEmail(),
            "firstname" => $user->getFirstname(),
            "lastname" => $user->getLastname(),
            "country" => $user->getCountry(),
            "phone" => $user->getPhone(),
            "city" => $user->getCity(),
            "company" => $user->getCompany(),
            "postalCode" => $user->getPostalCode(),
            "state" => $user->getState(),
            "street" => $user->getStreet(),
            "street2" => $user->getStreet2(),
            "street3" => $user->getStreet3(),
            "vatId" => $user->getVatId()
        );

        $return = $this->get('update-user', $data, false);

        return (!empty($return->success));
    }

    /**
     * @param int $imageId
     *
     * @return string
     */
    public function getImagePreview($imageId)
    {
        return $this->get("get-preview", array("objectId" => $imageId), true);
    }

    /**
     * @param string $keywords
     * @param int    $page
     * @param int    $itemsPerPage
     *
     * @return array
     */
    public function search($keywords, $page = 1, $itemsPerPage = null)
    {
        $params = array(
            'searchTerm' => $keywords,
            'page' => $page
        );

        if ($itemsPerPage !== null) {
            $params['itemsPerPage'] = $itemsPerPage;
        }

        $data = $this->get('search', $params, true);

        $items = array();
        foreach ($data->results as $result) {
            $item = new Item();
            $item->setId($result->id);
            $item->setIdExt($result->id_ext);
            $item->setTitle(mb_convert_case($result->title, MB_CASE_TITLE));

            $thumbnail = new Image();
            $thumbnail->setUrl($result->thumbnail_url);
            $thumbnail->setWidth($result->thumbnail_width);
            $thumbnail->setHeight($result->thumbnail_height);
            $item->setThumbnail($thumbnail);

            $preview = new Image();
            $preview->setUrl($result->preview_url);
            $preview->setWidth($result->preview_width);
            $preview->setHeight($result->preview_height);
            $item->setPreview($preview);

            $items[] = $item;
        }

        return array(
            "numberOfResults" => $data->number_of_results,
            "pageSize" => $data->page_size,
            "images" => $items
        );
    }

    /**
     * @param int $itemId
     * @param int $size
     *
     * @return \stdClass Object with properties `hash`, `type`, `file_size` and `checksum`.
     *
     * @throws ApiException
     */
    public function purchase($itemId, $size)
    {
        $params = array(
            "objectId" => $itemId,
            "size" => $size
        );

        return $this->get("purchase", $params, false);
    }

    /**
     * @param string $hash       Download hash
     * @param string $targetFile Filename including path where the new file should be downloaded to
     * @param string $checksum   MD5 checksum of the file to be downloaded
     *
     * @throws ApiException
     */
    public function download($hash, $targetFile, $checksum)
    {
        $ch = self::initCurl(self::PIXELROCKSTAR_API_URL . "download.json", array("hash" => $hash));

        $writeMode = "w";
        if (file_exists($targetFile)) {
            $writeMode = "a";
            curl_setopt($ch, CURLOPT_RANGE, filesize($targetFile) . "-");
        }
        $fp = fopen($targetFile, $writeMode);

        curl_setopt($ch, CURLOPT_FILE, $fp);
        curl_setopt($ch, CURLOPT_HEADERFUNCTION, array('\PixelRockstar\Api\Api', 'curlHeaderCallback'));

        $result = curl_exec($ch);
        curl_close($ch);
        fclose($fp);

        // Successful request returns `1`
        if ($result != 1) {
            unlink($targetFile);
            throw new ApiException(ApiException::IMAGE_DOWNLOAD_FAILED, ApiException::CODE_IMAGE_DOWNLOAD_FAILED);
        }

        if (md5_file($targetFile) != $checksum) {
            unlink($targetFile);
            throw new ApiException(
                ApiException::IMAGE_DOWNLOAD_BAD_CHECKSUM,
                ApiException::CODE_IMAGE_DOWNLOAD_BAD_CHECKSUM
            );
        }
    }

    /**
     * @param string $hash Download hash
     * @param string $type Any of `NOTIFY_FILE_DOWNLOAD_*` consts
     *
     * @return \stdClass
     */
    public function notify($hash, $type)
    {
        $params = array(
            "hash" => $hash,
            "type" => $type
        );

        return $this->get("notify", $params, false);
    }
}
