<?php

namespace PixelRockstar;

use PixelRockstar\Api\Api;
use PixelRockstar\Api\ApiException;
use PixelRockstar\Api\Credentials;
use PixelRockstar\Settings\Constants;
use PixelRockstar\Settings\Settings;
use PixelRockstar\User\User;

class PixelRockstar
{
    private static $initiated = false;

    /** @var bool If current user can manage plugins = is admin */
    private static $canManagePlugin = false;

    public static $languageDomain = "pixelrockstar";
    public static $appId = "pixelrockstar";

    /** @var User */
    public static $user;

    public static function init()
    {
        // Ignore WP's heartbeat requests
        if (getPostValue("action") === "heartbeat") {
            return;
        }

        // Do not initialize if already initialized or if we aren't in admin area
        if (self::$initiated || !is_admin()) {
            return;
        }

        self::$canManagePlugin = current_user_can("activate_plugins");

        self::$user = new User();
        // Try to log in
        try {
            $credentials = self::getCredentialsFromSettings();
            self::login($credentials);
        } catch (\Exception $e) {
        }

        self::initHooks();
    }

    private static function initHooks()
    {
        self::$initiated = true;

        if (self::$canManagePlugin) {
            add_action('wp_loaded', array('\PixelRockstar\PixelRockstarAdmin', 'init'), 100);
        }

        // Load text domain for translations
        load_plugin_textdomain(self::$languageDomain, false, dirname(PIXELROCKSTAR_PLUGIN_BASENAME) . '/languages/');

        wp_enqueue_style(
            self::$appId . '_styles',
            PIXELROCKSTAR_PLUGIN_URL . '/css/styles.css',
            array(),
            PIXELROCKSTAR_VERSION
        );

        wp_enqueue_script("jquery");
        wp_enqueue_script(
            PixelRockstar::$appId . '_scripts',
            PIXELROCKSTAR_PLUGIN_URL . '/js/scripts.js',
            array('media-views'),
            PIXELROCKSTAR_VERSION
        );

        // Include jQuery.justifiedGallery
        wp_enqueue_script(
            PixelRockstar::$appId . '_scripts_justifiedGallery',
            PIXELROCKSTAR_PLUGIN_URL . '/js/jquery.justifiedGallery.js',
            array("jquery"),
            PIXELROCKSTAR_VERSION
        );

        // Add button on Edit Page next to "Add Media" button
        add_filter('media_buttons_context', array('\PixelRockstar\PixelRockstar', 'mediaButtonsContext'));

        // Pass strings into JS for Media Upload lightbox
        wp_localize_script(PixelRockstar::$appId . '_scripts', 'pixelrockstar_vars', self::getJavascriptVariablesArray());
        wp_localize_script(PixelRockstar::$appId . '_scripts', 'pixelrockstar_l10n', PixelRockstar::mediaUploadStrings());

        // Add upload stock page
        add_action("admin_menu", array('\PixelRockstar\PixelRockstar', 'adminMenu'));

        // Add AJAX handler
        add_action(
            'wp_ajax_' . PixelRockstar::$appId . '_action',
            array('\PixelRockstar\PixelRockstar', 'ajaxCallback')
        );
    }

    /**
     * Media buttons above post editor
     *
     * @param string $context
     *
     * @return string
     */
    public static function mediaButtonsContext($context)
    {
        $context .=
            '<a href="#" class="button pixelrockstar js-' . self::$appId . '-add-stock-photo"><em class="pixelrockstar-Icon pixelrockstar-Icon-icon"></em> ' .
            __('Add Stock Photo', self::$languageDomain) .
            '</a>';

        return $context;
    }

    /**
     * Media upload lightbox - left menu
     *
     * @param array $tabs
     *
     * @return array
     */
    public static function mediaUploadTabs($tabs)
    {
        $tabs[self::$appId] = __("Add Stock Photo", self::$languageDomain);

        return $tabs;
    }

    /**
     * Localized strings to be passed to the JS
     *
     * @return array
     */
    public static function mediaUploadStrings()
    {
        return array(
            'link' =>
                '<em class="pixelrockstar-Icon pixelrockstar-Icon-icon"></em> ' .
                __('Add Stock Photo', self::$languageDomain),
            'title' =>
                '<em class="pixelrockstar-Icon pixelrockstar-Icon-icon pixelrockstar-Icon-icon--title"></em> ' .
                __('Stock Photos', self::$languageDomain),
            'title2' => __('Stock Photos', self::$languageDomain),
            'button_search' => __('Search', self::$languageDomain),
            'button_next' => __('Next', self::$languageDomain),
            'button_back' => __('Back', self::$languageDomain),
            'button_download' => __('Download', self::$languageDomain),
            'label_addAsFeatured' => __('Add as a Featured Image', self::$languageDomain),
            'label_setDefaultSize' => __('Stick to this size', self::$languageDomain),
            'text_editSizePresets' => __('Edit size presets', self::$languageDomain),
            'link_editSizePresets' => "options-general.php?page=" . PixelRockstar::$appId,
            'input_placeholder_search' => __('Search for photos, vectors & illustrations', self::$languageDomain),
            'dialog_error_title' => __("Error", self::$languageDomain),
            'progressbar_title' => __("Delivering", self::$languageDomain),
            'progressbar_desc_prepare' => __("Catching image", self::$languageDomain),
            'progressbar_desc_optimize' => __("Optimizing image for better SEO", self::$languageDomain),
            'progressbar_desc_upload' => __("Uploading image to your website", self::$languageDomain),
            'progressbar_desc_add' => __("Adding image to library", self::$languageDomain),
            'text_sizesInfo' => __("Smaller size improves loading time. Larger brings more quality.", self::$languageDomain),
            'text_imageId' => sprintf(
                // TRANSLATORS: Image ID
                __("Image %s", self::$languageDomain),
                '<%= item.id_ext %>'
            ),
            'text_featured_image_error' => __("Failed to set Featured Image", self::$languageDomain)
        );
    }

    /**
     * Populates array accessible via `pixelrockstar_vars` variable in JS
     *
     * @return array
     */
    private static function getJavascriptVariablesArray()
    {
        $sizes = self::getSizesArray();
        $mediaSizes = self::getMediaSizesArray();
        $defaultSize = $sizes["default_size"];
        $sizesWithLegend = array();

        // Display sizes selector for admins or if "always ask" is default option
        if (self::$canManagePlugin || (isset($sizes["default_size"]) && $sizes["default_size"] == Constants::DEFAULT_SIZE_ASK)) {
            if ($sizes["default_size"] == Constants::DEFAULT_SIZE_ASK) {
                $defaultSize = $sizes["hires_max_size"];
            }

            $sizesWithLegend = array(
                array(
                    "size" => (int)$sizes["hires_max_size"],
                    "size_formatted" => number_format_i18n((int)$sizes["hires_max_size"]),
                    "title" => __("Full size", self::$languageDomain)
                ),
                array(
                    "size" => (int)$sizes["page_width"],
                    "size_formatted" => number_format_i18n((int)$sizes["page_width"]),
                    "title" => __("Page width", self::$languageDomain)
                ),
            );

            if (isset($sizes["default_size_media"]) && isset($mediaSizes[$sizes["default_size_media"]])) {
                $mediaSize = (int)$mediaSizes[$sizes["default_size_media"]]["width"];
                $sizesWithLegend[] = array(
                    "size" => $mediaSize,
                    "size_formatted" => number_format_i18n($mediaSize),
                    "title" => $mediaSizes[$sizes["default_size_media"]]["title"]
                );
                if ($defaultSize == Constants::DEFAULT_SIZE_MEDIA) {
                    $defaultSize = $mediaSize;
                }
            }

            $sizesWithLegend[] = array(
                "size" => "custom",
                "title" => __("Custom size", self::$languageDomain)
            );
        }

        $subscriptionError = "";
        $subscription = self::$user->getSubscription();
        if ($subscription->getDaysUntilEnd() <= 0) {
            $subscriptionError = __("Your subscription plan has expired.", self::$languageDomain);
        } else if ($subscription->getDownloadsAvailable() <= 0) {
            $subscriptionError = __("No more downloads left in your subscription plan.", self::$languageDomain);
        }

        return array(
            "sizes" => $sizesWithLegend,
            "custom_size" => (int)$sizes["default_size_custom"],
            "default_size" => $defaultSize,
            "can_change_default" => self::$canManagePlugin,
            "image_alignment" => Settings::getSetting("image_alignment", "none"),
            "subscription_error" => $subscriptionError
        );
    }

    /**
     * Call:
     *
     * var data = {
     *   'action': 'pixelrockstar_search_action',
     *   'what': [required action]
     *   ...
     * };
     * jQuery.post(ajaxurl, data, function(response) { ... });
     */
    public static function ajaxCallback()
    {
        $return = array(
            "success" => false
        );

        try {
            switch (getGetValue("what", getPostValue("what"))) {
                case "search":
                    $searchTerm = getGetValue("searchTerm");
                    $page = (int)getGetValue("page", 1);
                    $resultsPerPage = Settings::getSetting("search_results", Constants::SEARCH_RESULTS_50);

                    $searchResults = self::search($searchTerm, $page, $resultsPerPage);
                    if (empty($searchResults["images"])) {
                        throw new \Exception(__("Sorry, we have no results for this keyword.", self::$languageDomain));
                    }

                    $pageSize = (!empty($searchResults["pageSize"])) ? $searchResults["pageSize"] : $resultsPerPage;
                    if ($pageSize == 0) {
                        $pageSize = Constants::SEARCH_RESULTS_50;
                    }

                    $return = array(
                        "success" => true,
                        "searchResults" => $searchResults["images"],
                        "currentPage" => (int)(empty($searchResults["images"])) ? 1 : $page,
                        "totalPages" => (int)ceil($searchResults["numberOfResults"] / $pageSize),
                        "imagesFrom" => (int)$resultsPerPage * ($page - 1) + 1,
                        "imagesTo" => (int)min($resultsPerPage * $page, $searchResults["numberOfResults"])
                    );
                    break;
                case "preview":
                    $imageId = getGetValue("imageId");
                    $return = array(
                        "success" => true,
                        "url" => self::getPreviewUrl($imageId)
                    );
                    break;
                case "purchase":
                    $imageId = getPostValue("imageId");
                    $size = (int)getPostValue("size");
                    $purchaseData = self::purchaseImage($imageId, $size);

                    $return = array(
                        "success" => true,
                        "hash" => $purchaseData->hash,
                        "fileSize" => (int)$purchaseData->file_size,
                        "checksum" => $purchaseData->checksum
                    );
                    break;
                case "download":
                    $hash = getGetValue("hash");
                    $checksum = getGetValue("checksum");
                    $title = getGetValue("title", time());

                    $file = self::getAvailableFilename($title, "jpg");

                    self::download($hash, $file, $checksum);
                    $return = array(
                        "success" => true,
                        "file" => $file
                    );

                    break;
                case "add-to-library":
                    $file = getPostValue("file");
                    $title = getPostValue("title");
                    $searchTerm = getPostValue("searchTerm");

                    if (!$file || !file_exists($file)) {
                        throw new \Exception(__("File doesn't exist.", self::$languageDomain));
                    }

                    $return = array(
                        "success" => true,
                        "attachmentId" => self::addImageToLibrary($file, $title, $searchTerm)
                    );

                    break;
                case "update-default-size":
                    self::setDefaultSize(getPostValue("size"), getPostValue("customSize", null));

                    // We don't really care about result... :-)
                    $return = array(
                        "success" => true
                    );
                    break;
                default:
                    $return["error"] = __("Unknown action.", self::$languageDomain);
                    break;
            }
        } catch (\Exception $e) {
            $return["error"] = $e->getMessage();
        }

        if ($return["success"]) {
            header("Content-type: application/json");
            die(json_encode($return));
        } else {
            header("Content-type: text/plain");
            header("HTTP/1.0 404 Not Found");

            $error = @$return["error"] ?: __("Unknown error.", self::$languageDomain);
            die($error);
        }
    }

    /**
     * Gets Api class instance using credentials stored in DB
     *
     * @return Api
     *
     * @throws \Exception
     */
    private static function getApiInstance()
    {
        $credentials = self::getCredentialsFromSettings();

        return new Api($credentials);
    }

    /**
     * @param Credentials $credentials
     *
     * @return User
     */
    protected static function login(Credentials $credentials)
    {
        $api = new Api($credentials);
        self::$user = $api->getUser();

        return self::$user;
    }

    /**
     * Sets `Credentials` to be used in API calls
     *
     * @throws \Exception
     */
    protected static function getCredentialsFromSettings()
    {
        $apiKey = Settings::getSetting("api_key");

        if (empty($apiKey)) {
            throw new \Exception(__("Missing API key.", self::$languageDomain));
        }

        return new Credentials(Credentials::API_KEY_USERNAME, $apiKey);
    }

    /**
     * @param string $searchTerm
     * @param int    $page
     * @param null   $resultsPerPage
     *
     * @return array
     * @throws \Exception
     */
    public static function search($searchTerm, $page = 1, $resultsPerPage = null)
    {
        if (empty($searchTerm)) {
            throw new \Exception(__("No search term was specified.", self::$languageDomain));
        }

        $api = self::getApiInstance();

        if ($resultsPerPage === null) {
            $resultsPerPage = Settings::getSetting("search_results", Constants::SEARCH_RESULTS_50);
        }

        return $api->search($searchTerm, $page, $resultsPerPage);
    }

    /**
     * @param int $imageId
     *
     * @return string
     * @throws \Exception
     */
    public static function getPreviewUrl($imageId)
    {
        if (empty($imageId)) {
            throw new \Exception(__("Invalid image ID", self::$languageDomain));
        }

        $api = self::getApiInstance();
        $previewUrl = $api->getImagePreview($imageId);

        if (empty($previewUrl)) {
            throw new \Exception(__("Failed to get image preview", self::$languageDomain));
        }

        return $previewUrl;
    }

    /**
     * @param int $imageId
     * @param int $size
     *
     * @return \stdClass
     * @throws \Exception
     */
    private static function purchaseImage($imageId, $size)
    {
        if (empty($imageId)) {
            throw new \Exception(__("Invalid image ID", self::$languageDomain));
        }
        if ($size <= 0) {
            throw new \Exception(__("Invalid image size", self::$languageDomain));
        }

        $api = self::getApiInstance();
        $purchaseData = $api->purchase($imageId, $size);
        if (empty($purchaseData->hash)) {
            throw new \Exception(__("Failed to purchase image", self::$languageDomain));
        }

        return $purchaseData;
    }

    /**
     * @param string $hash       Download hash
     * @param string $targetFile Filename including path where the new file should be downloaded to
     * @param string $checksum   MD5 checksum of the file to be downloaded
     *
     * @throws \Exception
     */
    private static function download($hash, $targetFile, $checksum)
    {
        if (empty($hash) || empty($targetFile) || empty($checksum)) {
            throw new \Exception(__("Malformed download request", self::$languageDomain));
        }

        $api = self::getApiInstance();
        try {
            $api->download($hash, $targetFile, $checksum);
            $api->notify($hash, Api::NOTIFY_FILE_DOWNLOAD_OK);
        } catch (ApiException $e) {
            if ($e->getCode() === ApiException::CODE_IMAGE_DOWNLOAD_BAD_CHECKSUM) {
                $type = Api::NOTIFY_FILE_DOWNLOAD_BAD_CHECKSUM;
            } else {
                $type = Api::NOTIFY_FILE_DOWNLOAD_FAILED;
            }
            $api->notify($hash, $type);

            throw new \Exception($e->getMessage());
        }
    }

    /**
     * Adds image to WP media library
     *
     * @param string $file       Filename including path where the file is located
     * @param string $title      Image title
     * @param string $searchTerm Search term
     *
     * @return int Attachment ID
     */
    private static function addImageToLibrary($file, $title, $searchTerm = "")
    {
        $altText = "";
        $caption = "";

        $altTextType = Settings::getSetting("image_alt_text", Constants::DEFAULT_ALT_TEXT);
        $captionType = Settings::getSetting("caption_text", Constants::DEFAULT_CAPTION);

        switch ($altTextType) {
            case Constants::ALT_TEXT_TITLE:
                $altText = $title;
                break;
            case Constants::ALT_TEXT_KEYWORDS:
                $altText = $searchTerm ? $searchTerm : $title;
                break;
        }

        switch ($captionType) {
            case Constants::CAPTION_TEXT_COPYRIGHT:
                $caption = self::getCopyrightMessage();
                break;
            case Constants::CAPTION_TEXT_CUSTOM:
                $caption = Settings::getSetting("caption_text_custom");
                if (Settings::getSetting("caption_text_with_copyright", Constants::CAPTION_TEXT_CUSTOM_WITH_COPYRIGHT) == Constants::CAPTION_TEXT_CUSTOM_WITH_COPYRIGHT) {
                    if (!empty($caption)) {
                        $caption .= " &mdash; ";
                    }
                    $caption .= self::getCopyrightMessage();
                }
                break;
            case Constants::CAPTION_TEXT_IMAGE_TITLE:
                $caption = htmlentities2(mb_convert_case($title, MB_CASE_TITLE));
                break;
        }

        $wpFiletype = wp_check_filetype($file);
        $attachment = array(
            'guid' => $file,
            'post_mime_type' => $wpFiletype['type'],
            'post_title' => mb_convert_case($title, MB_CASE_TITLE),
            'post_content' => '',
            'post_excerpt' => $caption,
            'post_status' => 'inherit'
        );

        require_once(ABSPATH . 'wp-admin/includes/image.php');

        $attachId = wp_insert_attachment($attachment, $file);
        $attachData = wp_generate_attachment_metadata($attachId, $file);
        wp_update_attachment_metadata($attachId, $attachData);

        if ($altText) {
            update_metadata(
                'post',
                $attachId,
                '_wp_attachment_image_alt',
                htmlentities2(mb_convert_case($altText, MB_CASE_TITLE))
            );
        }

        return $attachId;
    }

    /**
     * Returns available filename including path in WP upload directory
     *
     * @param string $preferredFilename
     * @param string $extension
     *
     * @return string Filename including path
     */
    private static function getAvailableFilename($preferredFilename, $extension = "jpg")
    {
        $uploadDir = wp_upload_dir();
        $filename = $preferredFilename . "." . $extension;
        $file = $uploadDir["path"] . "/" . wp_unique_filename($uploadDir["path"], $filename);

        // "Reserve" the filename
        file_put_contents($file, "");

        return $file;
    }

    /**
     * @return array
     */
    public static function getSizesArray()
    {
        global $content_width;

        return array(
            "default_size" => Settings::getSetting("default_size", Constants::DEFAULT_SIZE_ASK),
            "hires_max_size" => Constants::MAX_HIRES_SIZE,
            "page_width" => $content_width,
            "default_size_custom" => Settings::getSetting("default_size_custom"),
            "default_size_media" => Settings::getSetting("default_size_media"),
        );
    }

    public static function getMediaSizesArray()
    {
        return array(
            "thumbnail" => array(
                "title" => __("Thumbnail size"),
                "width" => get_option('thumbnail_size_w')
            ),
            "medium" => array(
                "title" => __("Medium size"),
                "width" => get_option('medium_size_w')
            ),
            "large" => array(
                "title" => __("Large size"),
                "width" => get_option('large_size_w')
            )
        );
    }

    /**
     * @param string $selectedSize
     * @param int    $customSize Custom size if `$selectedSize == Constants::DEFAULT_SIZE_CUSTOM`
     *
     * @return bool
     * @throws \Exception
     */
    private static function setDefaultSize($selectedSize, $customSize = null)
    {
        if (!self::$canManagePlugin) {
            throw new \Exception(__("You are not allowed to change settings.", self::$languageDomain));
        }

        $customSize = (int)$customSize;
        $settings = array(
            "default_size" => $selectedSize
        );

        if ($selectedSize == Constants::DEFAULT_SIZE_CUSTOM && $customSize > 0 && $customSize <= Constants::MAX_HIRES_SIZE) {
            $settings["default_size_custom"] = $customSize;
        }

        return Settings::setSettings($settings);
    }

    public static function getCopyrightMessage()
    {
        return __("© PixelRockstar.com", self::$languageDomain);
    }

    /**
     * Gets list of purchased files
     *
     * @return array
     */
    public static function getDownloadedFiles()
    {
        global $wpdb;

        $rows = $wpdb->get_results("
            SELECT
              prs.meta_value AS prs_meta,
              image.meta_value AS filename
            FROM {$wpdb->prefix}postmeta AS prs
            JOIN {$wpdb->prefix}postmeta AS image ON prs.post_id = image.post_id 
            WHERE
              prs.meta_key = '" . PIXELROCKSTAR_IMAGE_META_KEY . "' AND
              image.meta_key = '_wp_attached_file'
        ", ARRAY_A);

        $images = array();
        foreach ($rows as $row) {
            $prsMeta = unserialize($row['prs_meta']);
            if ($prsMeta['name'] != PIXELROCKSTAR_PLUGIN_NAME || isset($prsMeta['aviary_edited'])) {
                continue;
            }

            $images[] = $row['filename'];
        }

        return $images;
    }

    public static function adminMenu()
    {
        // Add menu in Settings
        add_media_page(
            __("Add Stock Photo", PixelRockstar::$languageDomain),
            __("Add Stock Photo", PixelRockstar::$languageDomain),
            "upload_files",
            PixelRockstar::$appId . "-upload",
            array('\PixelRockstar\PixelRockstar', 'renderUploadPage')
        );
    }

    public static function renderUploadPage()
    {
        $strings = array('pixelrockstar' => PixelRockstar::mediaUploadStrings());

        wp_localize_script('media-views', '_wpMediaViewsL10n', $strings);
        wp_enqueue_style('media-views');

        renderPage("upload.php", $strings);
    }
}
