<?php

namespace PixelRockstar;

use PixelRockstar\Api\Api;
use PixelRockstar\Api\Credentials;
use PixelRockstar\Country\CountriesProvider;
use PixelRockstar\Settings\Constants;
use PixelRockstar\Settings\Settings;

class PixelRockstarAdmin
{
    private static $initiated = false;

    public static function init()
    {
        if (self::$initiated) {
            return;
        }

        self::initHooks();
    }

    public static function initHooks()
    {
        self::$initiated = true;

        // Add option page
        add_action("admin_menu", array('\PixelRockstar\PixelRockstarAdmin', 'adminMenu'));

        // Add AJAX handler
        add_action(
            'wp_ajax_' . PixelRockstar::$appId . "_admin_action",
            array('\PixelRockstar\PixelRockstarAdmin', 'adminAjaxCallback')
        );

        wp_enqueue_script('jquery');
    }

    /**
     * Call:
     *
     * var data = {
     *   'action': 'pixelrockstar_admin_action',
     *   'what': [required action]
     *   ...
     * };
     * jQuery.post(ajaxurl, data, function(response) { ... });
     */
    public static function adminAjaxCallback()
    {
        $return = array(
            "success" => false
        );

        try {
            switch (getPostValue("what")) {
                case "login":
                    $apiKey = getPostValue("apikey");
                    if (strlen($apiKey) != 40) {
                        throw new \Exception(__("Invalid API key.", PixelRockstar::$languageDomain));
                    }

                    $credentials = new Credentials(Credentials::API_KEY_USERNAME, $apiKey);
                    $api = new Api($credentials);
                    $user = $api->getUser();
                    if ($user->getUserId() == "") {
                        throw new \Exception(__("Invalid API key.", PixelRockstar::$languageDomain));
                    }

                    self::flushLoginData();

                    $settings = array(
                        "username" => $user->getUsername(),
                        "api_key" => $credentials->getPassword()
                    );
                    if (Settings::setSettings($settings)) {
                        $return["success"] = true;
                        $return["username"] = $user->getUsername();
                        $return["downloads_remaining"] = $user->getSubscription()->getDownloadsAvailable();
                        $return["days_remaining"] = $user->getSubscription()->getDaysUntilEnd();
                    } else {
                        $return["error"] = __("Failed to save API key.", PixelRockstar::$languageDomain);
                    }
                    break;
                case "logout":
                    if (self::flushLoginData()) {
                        $return["success"] = true;
                    } else {
                        $return["error"] = __("Failed to log out.", PixelRockstar::$languageDomain);
                    }
                    break;
                case "update-settings":
                    $dataRaw = (isset($_POST["data"])) ? ($_POST["data"]) : null;
                    parse_str($dataRaw, $data);
                    if (empty($data)) {
                        $return["error"] = __("Nothing to save.", PixelRockstar::$languageDomain);
                        break;
                    }

                    foreach ($data as $key => $value) {
                        $value = str_replace("\\'", "'", $value);
                        $data[$key] = sanitize_text_field($value);
                    }

                    if (self::saveSettings($data)) {
                        $return["success"] = true;
                        $return["message"] = __("Your settings were just saved.", PixelRockstar::$languageDomain);
                    } else {
                        $return["error"] = __("Failed to save settings.", PixelRockstar::$languageDomain);
                    }
                    break;
                default:
                    $return["error"] = __("Unknown action.", PixelRockstar::$languageDomain);
                    break;
            }
        } catch (\Exception $e) {
            $return["error"] = $e->getMessage();
        }

        die(json_encode($return));
    }

    public static function adminMenu()
    {
        // Add menu in Settings
        add_options_page(
            __("PixelRockstar", PixelRockstar::$languageDomain),
            __("PixelRockstar", PixelRockstar::$languageDomain),
            "manage_options",
            PixelRockstar::$appId,
            array('\PixelRockstar\PixelRockstarAdmin', 'renderSettingsPage')
        );
        // Add custom links in plugin settings
        add_filter(
            "plugin_action_links_" . plugin_basename(PIXELROCKSTAR_PLUGIN_BASENAME),
            array('\PixelRockstar\PixelRockstarAdmin', 'filterPluginActions')
        );
    }

    public static function filterPluginActions($links)
    {
        $links[] = '<a href="options-general.php?page=' . PixelRockstar::$appId . '">' . __('Settings') . '</a>';

        return $links;
    }

    public static function renderSettingsPage()
    {
        wp_enqueue_script(
            PixelRockstar::$appId . '_admin_settings',
            PIXELROCKSTAR_PLUGIN_URL . '/js/admin_settings.js',
            array(),
            PIXELROCKSTAR_VERSION
        );

        $params = array(
            "user" => array(
                "logged_in" => (PixelRockstar::$user->getUserId() > 0),
                "username" => Settings::getSetting("username"),
                "email" => PixelRockstar::$user->getEmail(),
                "firstname" => PixelRockstar::$user->getFirstname(),
                "lastname" => PixelRockstar::$user->getLastname(),
                "company" => PixelRockstar::$user->getCompany(),
                "country" => PixelRockstar::$user->getCountry(),
                "countries" => CountriesProvider::getCountries(),
            ),
            "size" => PixelRockstar::getSizesArray(),
            "mediaSizes" => PixelRockstar::getMediaSizesArray(),
            "image_alignment" => Settings::getSetting("image_alignment"),
            "search_results" => Settings::getSetting("search_results", Constants::SEARCH_RESULTS_50),
            "caption" => array(
                "text" => Settings::getSetting("caption_text", Constants::DEFAULT_CAPTION),
                "custom_text" => Settings::getSetting("caption_text_custom"),
                "custom_with_copyright" => Settings::getSetting("caption_text_with_copyright", Constants::CAPTION_TEXT_CUSTOM_WITH_COPYRIGHT)
            ),
            "alt_text" => Settings::getSetting("image_alt_text", Constants::DEFAULT_ALT_TEXT),
            "url" => array(
                "contact_email" => PIXELROCKSTAR_CONTACT_EMAIL,
                "facebook" => PIXELROCKSTAR_FACEBOOK_URL,
                "twitter" => PIXELROCKSTAR_TWITTER_URL,
                "rate_us" => PIXELROCKSTAR_RATE_US_URL,
                "license_agreement" => PIXELROCKSTAR_LICENSE_URL,
                "terms_of_sercive" => PIXELROCKSTAR_TERMS_OF_SERVICE_URL,
                "privacy_policy" => PIXELROCKSTAR_PRIVACY_POLICY_URL,
                "buy_downloads" => "https://www.pixelrockstar.com/#PRICING",
                "create_account" => "https://www.pixelrockstar.com/my-account/",
                "forgotten_api_key" => "https://www.pixelrockstar.com/my-account/",
                "update_credentials" => "https://www.pixelrockstar.com/my-account/",
                "cancel_renewal" => "https://www.pixelrockstar.com/my-account/",
            ),
            "subscription" => array(
                "downloads_remaining" => PixelRockstar::$user->getSubscription()->getDownloadsAvailable(),
                "days_remaining" => PixelRockstar::$user->getSubscription()->getDaysUntilEnd(),
                "package_name" => PixelRockstar::$user->getSubscription()->getPackageName(),
                "renewal_date" => PixelRockstar::$user->getSubscription()->getEndDateFormatted()
            ),
            "images_downloaded" => count(PixelRockstar::getDownloadedFiles())
        );
        renderPage("settings.php", $params);
    }

    /**
     * @return bool
     */
    private static function flushLoginData()
    {
        $remove = array(
            "username",
            "password",
            "api_key"
        );

        return Settings::removeSettings($remove);
    }

    /**
     * @param array $settings
     *
     * @return bool
     */
    private static function saveSettings(array $settings)
    {
        $allowedSettings = array(
            "default_size",
            "default_size_custom",
            "default_size_media",
            "image_alignment",
            "search_results",
            "caption_text",
            "caption_text_custom",
            "caption_text_with_copyright",
            "image_alt_text"
        );

        // If updating Advanced Settings set `false` to unchecked checkbox
        if (isset($settings["caption_text"]) && !isset($settings["caption_text_with_copyright"])) {
            $settings["caption_text_with_copyright"] = false;
        }

        foreach ($settings as $key => $value) {
            if (!in_array($key, $allowedSettings)) {
                unset($settings[$key]);
                continue;
            }

            if ($key == "default_size_media" && (int)$value > Constants::MAX_HIRES_SIZE) {
                unset($settings[$key]);
                continue;
            }
        }

        if (empty($settings)) {
            return false;
        }

        return Settings::setSettings($settings);
    }
}